from logging import getLogger
from typing import Any, Dict, List, Optional, Type, TypeVar

from draccus.utils import get_type_arguments, is_optional, is_tuple, is_union

T = TypeVar("T")

logger = getLogger(__name__)

_new_metavars: Dict[Type, Optional[str]] = {
    # the 'primitive' types don't get a 'new' metavar.
    t: t.__name__
    for t in [str, float, int, bytes]
}


def get_metavar(t: Type, top_level: bool = True) -> Optional[str]:
    """Gets the metavar to be used for that type in help strings.

    This is crucial when using a `weird` auto-generated parsing functions for
    things like Union, Optional, Etc etc.

    type the type arguments that were passed to `get_parsing_fn` that
    produced the given parsing_fn.

    returns None if the name shouldn't be changed.
    """
    # TODO: Maybe we can create the name for each returned call, a bit like how
    # we dynamically create the parsing function itself?
    new_name: Optional[str] = getattr(t, "__name__", None)

    optional = is_optional(t)

    if t in _new_metavars:
        return _new_metavars[t]

    elif is_union(t):
        args = get_type_arguments(t)
        metavars: List[str] = []
        for type_arg in args:
            if type_arg is type(None):
                if top_level:
                    continue
                else:
                    metavars.append("None")
            else:
                metavars.append(get_metavar(type_arg) or "<unknown>")
        metavar = "|".join(map(str, metavars))
        if optional and top_level:
            return f"[{metavar}]"
        return metavar

    elif is_tuple(t):
        args = get_type_arguments(t)
        if not args:
            return get_metavar(Any)  # type: ignore
        logger.debug(f"Tuple args: {args}")
        metavars = []
        for arg in args:
            if arg is Ellipsis:
                metavars.append(f"[{metavars[-1]}, ...]")
                break
            else:
                metavars.append(get_metavar(arg) or "<unknown>")
        return " ".join(metavars)

    else:
        try:
            args = get_type_arguments(t)
            if args:
                # we want the name without type args
                # List[str]'s __name__ is List[str], but we want List
                new_name = t.__origin__.__name__
                return f"{new_name}[{','.join(get_metavar(arg, top_level=False) or str(arg) for arg in args)}]"
        except Exception:
            pass

    return new_name



""" Functions for decoding dataclass fields from "raw" values (e.g. from json).
"""
import functools
import typing
from collections import OrderedDict
from dataclasses import MISSING, fields, is_dataclass
from functools import lru_cache, partial
from logging import getLogger
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Protocol,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_args,
)

from draccus.choice_types import CHOICE_TYPE_KEY, ChoiceType
from draccus.parsers.registry_utils import RegistryFunc, withregistry
from draccus.utils import (
    DecodingError,
    ParsingError,
    canonicalize_union,
    format_error,
    get_type_arguments,
    has_generic_arg,
    is_choice_type,
    is_dict,
    is_enum,
    is_list,
    is_literal,
    is_set,
    is_tuple,
    is_union,
    stringify_type,
)

logger = getLogger(__name__)

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
K = TypeVar("K")
V = TypeVar("V")
Dataclass = TypeVar("Dataclass")


class DecodingFunction(Protocol[T_co]):
    def __call__(self, raw_value: Any, path: Sequence[str]) -> T_co:
        ...


@withregistry
def decode(cls: Type[T], raw_value: Any) -> T:
    cls = canonicalize_union(cls)
    return get_decoding_fn(cls)(raw_value, ())  # type: ignore


def decode_from_init(cls: Type[T], raw_value: Any, path: Sequence[str]) -> T:
    """Decodes a value into an atomic type (e.g. str, int, float, etc.)."""
    try:
        return cls(raw_value)  # type: ignore
    except Exception as e:
        raise DecodingError(path, f"Couldn't parse '{raw_value}' into a {stringify_type(cls)}") from e


for t in [str, float, bytes]:
    decode.register(t, partial(decode_from_init, t))


def apply_type_map(typ: Any, type_map: Dict[TypeVar, Type]) -> Any:
    """Recursively replaces TypeVars in container types like List[T], Dict[K, V]"""
    if isinstance(typ, TypeVar):
        return type_map.get(typ, typ)
    origin = typing.get_origin(typ)
    if origin is not None:
        args = tuple(apply_type_map(arg, type_map) for arg in typing.get_args(typ))
        if is_union(typ):  # union's origin doesn't like args
            return Union[args]
        if len(args) == 1:
            return origin[args[0]]  # ClassVar doesn't like tuples
        elif len(args) == 0:
            return origin

        return origin[args]
    return typ


@decode.register(bool)
def decode_bool(raw_value: Any, path) -> bool:
    # only accept yaml 1.2 bools
    if raw_value is True or raw_value == "true":
        return True
    elif raw_value is False or raw_value == "false":
        return False
    else:
        raise DecodingError(path, f"Couldn't parse '{raw_value}' into a bool")


@decode.register(int)
def decode_int(raw_value: Any, path) -> int:
    try:
        # reject floats etc:
        if isinstance(raw_value, float):
            raise ValueError(f"Expected an int, got a float: {raw_value}")
        return int(raw_value)
    except ValueError as e:
        raise DecodingError(path, f"Couldn't parse '{raw_value}' into an int") from e


def is_generic_type(cls: Type[Any]) -> bool:
    return typing.get_origin(cls) is not None


def decode_dataclass(cls: Type[Dataclass], d: Dict[str, Any], path: Sequence[str] = ()) -> Dataclass:
    path = tuple(path)
    obj_dict: Dict[str, Any] = d.copy()
    init_args: Dict[str, Any] = {}
    non_init_args: Dict[str, Any] = {}
    logger.debug(f"from_dict for {cls}")

    origin = typing.get_origin(cls)
    if origin is not None:
        type_args = typing.get_args(cls)
        type_vars = origin.__parameters__
        type_map = dict(zip(type_vars, type_args))
    else:
        origin = cls
        type_map = {}

    hints = {name: apply_type_map(t, type_map) for name, t in typing.get_type_hints(origin).items()}

    for field in fields(origin):
        name = field.name
        field_type = hints.get(name, field.type)
        if name not in obj_dict:
            continue
        raw_value = obj_dict.pop(name)
        try:
            logger.debug(f"Decode name = {name}, type = {field_type}")
            field_value = get_decoding_fn(field_type)(raw_value, (*path, name))
        except (ParsingError, DecodingError) as e:
            raise e
        except Exception as e:
            raise DecodingError(
                (*path, name),
                f"Failed when parsing value='{raw_value}' into field \"{cls}.{name}\" of type"
                f' {field_type}.\n\tUnderlying error is "{format_error(e)}"',
            ) from e
        if field.init:
            init_args[name] = field_value
        else:
            non_init_args[name] = field_value
    extra_args = obj_dict
    if extra_args:
        formatted_keys = ", ".join(f"`{k}`" for k in extra_args.keys())
        raise DecodingError(path, f"The fields {formatted_keys} are not valid for {stringify_type(cls)}")
    missing_fields = [
        field.name
        for field in fields(origin)
        if field.init and field.name not in init_args and field.default is MISSING and field.default_factory is MISSING
    ]
    if missing_fields:
        formatted_keys = ", ".join(f"`{k}`" for k in missing_fields)
        raise DecodingError(path, f"Missing required field(s) {formatted_keys} for {stringify_type(cls)}")
    try:
        instance = origin(**init_args)
    except (TypeError, ValueError) as e:
        raise ParsingError(f"Couldn't instantiate class {stringify_type(cls)} using the given arguments.") from e
    for name, value in non_init_args.items():
        logger.debug(f"Setting non-init field '{name}' on the instance.")
        setattr(instance, name, value)
    return instance


def decode_choice_class(cls: Type[T], raw_value: Any, path: Sequence[str]) -> T:
    """Decodes a value into an subtype of a choice class following the ChoiceType protocol."""
    assert issubclass(cls, ChoiceType)

    try:
        cls.get_choice_name(cls)
        # we already know what type we're looking for, so we can just use that
        return decode_dataclass(cls, raw_value)  # type: ignore
    except ValueError:
        pass

    if not isinstance(raw_value, dict):
        raise ParsingError(f"Expected a dict for a choice class, got {raw_value}")

    if CHOICE_TYPE_KEY not in raw_value:
        default = cls.default_choice_name()
        if default is None:
            raise ParsingError(f"Expected a dict with a '{CHOICE_TYPE_KEY}' key for {cls}, got {raw_value}")

        choice_type = default
    else:
        choice_type = raw_value[CHOICE_TYPE_KEY]

    try:
        subcls = cls.get_choice_class(choice_type)
    except KeyError as e:
        raise DecodingError(path, f"Couldn't find a choice class for '{choice_type}' in {cls}") from e

    raw_value = raw_value.copy()
    if CHOICE_TYPE_KEY in raw_value:
        raw_value.pop(CHOICE_TYPE_KEY)

    # return decode(subcls, raw_value)
    return decode_dataclass(subcls, raw_value, path)


def has_custom_decoder(cls: Type[T]):
    cached_func: RegistryFunc = decode.dispatch(cls)

    return cached_func is not None


@lru_cache(maxsize=100)
def get_decoding_fn(cls: Type[T]) -> DecodingFunction[T]:
    """Fetches/Creates a decoding function for the given type annotation.

    This decoding function can then be used to create an instance of the type
    when deserializing dicts.

    This function inspects the type annotation and creates the right decoding
    function recursively in a "dynamic-programming-ish" fashion.
    NOTE: We cache the results in a `functools.lru_cache` decorator to avoid
    wasteful calls to the function. This makes this process pretty efficient.

    """
    # Start by trying the dispatch mechanism
    underlying_type = typing.get_origin(cls) or cls
    cached_func: RegistryFunc = decode.dispatch(cls) or decode.dispatch(underlying_type)

    if cached_func is not None:
        # If supports subclasses, pass the actual type
        if cached_func.include_subclasses:
            return partial(cached_func.func, cls)
        else:
            fn = cached_func.func

            # we want to support the old interface where the decoding function
            # takes only one argument, so we wrap it here
            @functools.wraps(fn)
            def backwards_compat_call(raw_value: Any, path: Sequence[str] = ()) -> T:
                try:
                    return fn(raw_value, path)
                except TypeError:
                    try:
                        return fn(raw_value)
                    except Exception as e:  # pylint: disable=broad-except
                        raise DecodingError(
                            path, f"Couldn't parse '{raw_value}' into a {stringify_type(cls)}: {e}"
                        ) from e

            return backwards_compat_call

    elif is_choice_type(underlying_type):
        return partial(decode_choice_class, cls)

    elif is_dataclass(underlying_type):
        return partial(decode_dataclass, cls)

    elif cls is Any:
        logger.debug(f"Decoding an Any type: {cls}")
        return no_op

    elif is_dict(underlying_type):
        logger.debug(f"Decoding a Dict field: {cls}")
        args = get_type_arguments(cls)
        if args is None or len(args) != 2 or has_generic_arg(args):
            args = (Any, Any)
        return decode_dict(*args)

    elif is_set(cls):
        logger.debug(f"Decoding a Set field: {cls}")
        args = get_type_arguments(cls)
        if args is None or len(args) != 1 or has_generic_arg(args):
            args = (Any,)
        return decode_set(args[0])

    elif is_tuple(cls):
        logger.debug(f"Decoding a Tuple field: {cls}")
        args = get_type_arguments(cls)
        if args is None:
            args = []
        return decode_tuple(*args)

    elif is_list(cls):  # NOTE: Looks like can't be written with a dictionary
        logger.debug(f"Decoding a List field: {cls}")
        args = get_type_arguments(cls)
        if args is None or len(args) != 1 or has_generic_arg(args):
            # Using a `List` or `list` annotation, so we don't know what do decode the
            # items into!
            args = (Any,)
        decode_fn = decode_list(args[0])

        return decode_fn

    elif is_union(cls):
        logger.debug(f"Decoding a Union field: {cls}")
        args = get_type_arguments(cls)
        return decode_union(*args)

    elif is_enum(cls):
        return partial(decode_enum, cls)

    elif is_literal(cls):
        return partial(decode_literal, cls)

    import typing_inspect as tpi

    if tpi.is_typevar(cls):
        bound = tpi.get_bound(cls)
        logger.debug(f"Decoding a typevar: {cls}, bound type is {bound}.")
        if bound is not None:
            return get_decoding_fn(bound)

    raise Exception(f"No decoding function for type {cls}, consider using draccus.decode.register")


def decode_enum(cls: Type[T], raw_value: Any, path) -> T:
    """Decodes a value into an enum."""
    if not is_enum(cls):
        raise Exception(f"Expected an enum type, got {cls}")

    try:
        return cls(raw_value)  # type: ignore
    except ValueError:
        try:
            return cls[raw_value]  # type: ignore
        except ValueError as e:
            raise DecodingError(path, f"Couldn't parse '{raw_value}' into an enum of type {cls}") from e


def decode_optional(t: Type[T]) -> DecodingFunction[Optional[T]]:
    decode = get_decoding_fn(t)  # type: ignore

    def _decode_optional(raw_value: Any, path: Sequence[str] = ()) -> Optional[T]:
        return raw_value if raw_value is None else decode(raw_value, path)

    return _decode_optional


@typing.no_type_check
def decode_union(*types: Type[T]) -> DecodingFunction[T]:
    types = list(types)
    is_optional = type(None) in types
    # Partition the Union into None and non-None types.
    while type(None) in types:
        types.remove(type(None))

    decoding_fns = {t: get_decoding_fn(t) for t in types}
    # Try using each of the non-None types, in succession

    if len(decoding_fns) == 0:
        raise ValueError("Must provide at least one function to try")
    elif len(decoding_fns) == 1 and not is_optional:
        return next(iter(decoding_fns.values()))

    def _try_functions(val: Any, path: Sequence[str] = ()) -> T:
        if is_optional and val is None:
            return None

        exceptions = {}
        for descriptor, func in decoding_fns.items():
            try:
                return func(val, path)
            except Exception as e:
                exceptions[descriptor] = e

        message = "Could not decode the value into any of the given types:\n"
        for descriptor, ex in exceptions.items():
            if isinstance(ex, DecodingError):
                ex = ex.strip_prefix(path)
            descriptor = stringify_type(descriptor)
            submessage = str(ex).split("\n")
            # indent the submessage by (4 + len(descriptor) + 1) spaces
            first_line = True
            for line in submessage:
                if first_line:
                    first_line = False
                    message += f"    {descriptor}: {line.strip()}"
                    message += "\n"
                else:
                    message += f"{' ' * (5 + len(str(descriptor)))}{line}\n"

        exception_to_raise_from = next(iter(exceptions.values()))
        # we'd prefer to find the first exception with an interesting message (not a DecodingError)
        for ex in exceptions.values():
            if not isinstance(ex, DecodingError):
                exception_to_raise_from = ex
                break

        raise DecodingError(path, message) from exception_to_raise_from

    return _try_functions


def decode_list(t: Type[T]) -> DecodingFunction[List[T]]:
    decode_item = get_decoding_fn(t)  # type: ignore

    def _decode_list(raw_value: List[Any], path: Sequence[str]) -> List[T]:
        path = tuple(path)
        # assert type(val) == list
        if not isinstance(raw_value, list):
            raise Exception(f"The given value='{raw_value}' is not of a valid input for a list type")
        return [decode_item(v, (*path, str(i))) for i, v in enumerate(raw_value)]

    return _decode_list


def decode_tuple(*tuple_item_types: Type[T]) -> DecodingFunction[Tuple[T, ...]]:
    """Makes a parsing function for creating tuples."""
    # Get the decoding function for each item type
    has_ellipsis = False
    if Ellipsis in tuple_item_types:
        ellipsis_index = tuple_item_types.index(Ellipsis)
        decoding_fn_index = ellipsis_index - 1
        decoding_fn = get_decoding_fn(tuple_item_types[decoding_fn_index])  # type: ignore
        has_ellipsis = True
    elif len(tuple_item_types) == 0:
        has_ellipsis = True
        decoding_fn = no_op  # Functionality will be the same as Tuple[Any,...]
    else:
        decoding_fns = [get_decoding_fn(t) for t in tuple_item_types]  # type: ignore

    # Note, if there are more values than types in the tuple type, then the
    # last type is used.

    def _decode_tuple(raw_value: typing.Sequence[Any], path) -> Tuple[T, ...]:
        path = tuple(path)
        if raw_value is None:
            raise DecodingError("Value must not be None for conversion to a tuple", path)
        if has_ellipsis:
            return tuple(decoding_fn(v, (*path, str(i))) for i, v in enumerate(raw_value))
        else:
            if len(decoding_fns) != len(raw_value):
                # err_msg = f"Trying to decode {len(raw_value)} values for a predfined {len(decoding_fns)}-Tuple"
                err_msg = f"Expected {len(decoding_fns)} items, got {len(raw_value)}"
                raise DecodingError(path, err_msg)
            return tuple(decoding_fns[i](v, (*path, str(i))) for i, v in enumerate(raw_value))

    return _decode_tuple


def decode_set(item_type: Type[T]) -> DecodingFunction[Set[T]]:
    """Makes a parsing function for creating sets with items of type `item_type`."""
    # Get the parsers fn for a list of items of type `item_type`.
    parse_list_fn = decode_list(item_type)

    def _decode_set(raw_value: List[Any], path) -> Set[T]:
        return set(parse_list_fn(raw_value, path))

    return _decode_set


def decode_dict(K_: Type[K], V_: Type[V]) -> DecodingFunction[Dict[K, V]]:
    """Creates a decoding function for a dict type. Works with OrderedDict too."""
    decode_k = get_decoding_fn(K_)  # type: ignore
    decode_v: DecodingFunction[V] = get_decoding_fn(V_)  # type: ignore

    def _decode_dict(raw_value: Union[Dict[Any, Any], List[Tuple[Any, Any]]], path) -> Dict[K, V]:
        result: Dict[K, V] = {}
        items: Iterable[Tuple[Any, Any]]
        if isinstance(raw_value, list):
            result = OrderedDict()
            items = raw_value
        elif isinstance(raw_value, OrderedDict):
            # NOTE(ycho): Needed to propagate `OrderedDict` type
            result = OrderedDict()
            items = raw_value.items()
        else:
            items = raw_value.items()
        for k, v in items:
            k_ = decode_k(k, (*tuple(path), f"key={k}"))
            v_ = decode_v(v, (*tuple(path), k))
            result[k_] = v_
        return result

    return _decode_dict


def no_op(raw_value: T, path) -> T:
    """Decoding function that gives back the value as-is."""
    del path
    return raw_value


decode.register(Path, partial(decode_from_init, Path))


def decode_literal(cls: Type[T], raw_value: Any, path) -> T:
    """Decodes a value into a literal type."""
    allowed_values = get_args(cls)
    logger.debug(f"Decoding literal value '{raw_value}' ({type(raw_value)}) into one of: {allowed_values}")

    # First try direct equality with type checking
    for value in allowed_values:
        if raw_value == value and type(raw_value) == type(value):
            logger.debug(f"Found direct match with value {value}")
            return value

    # Then try type conversion for string inputs
    if isinstance(raw_value, str):
        logger.debug(f"Attempting type conversion for string input '{raw_value}'")
        # Try string comparison for string literals first
        for value in allowed_values:
            if isinstance(value, str):
                if raw_value == value:
                    logger.debug(f"Found string match with value {value}")
                    return value  # type: ignore

        # Try numeric conversion - only accept strings that look like numbers
        if raw_value.replace(".", "").replace("-", "").isdigit():
            for value in allowed_values:
                if isinstance(value, (int, float)):
                    try:
                        converted = type(value)(raw_value)
                        if converted == value:
                            logger.debug(f"Successfully converted to numeric value {value}")
                            return value  # type: ignore
                    except (ValueError, TypeError):
                        continue

        # Try boolean conversion - only accept "true" or "false"
        if raw_value.lower() in ("true", "false"):
            converted = raw_value.lower() == "true"
            for value in allowed_values:
                if isinstance(value, bool) and converted == value:
                    logger.debug(f"Successfully converted to boolean value {value}")
                    return value  # type: ignore

    # If nothing worked, raise an error with a descriptive message
    error_msg = (
        f"Cannot convert '{raw_value}' ({type(raw_value)}) into one of: {', '.join(str(v) for v in allowed_values)}"
    )
    logger.debug(f"Raising error: {error_msg}")
    raise DecodingError(path, error_msg)



from dataclasses import dataclass
from enum import Enum, auto
from typing import Dict, Generic, Literal, Tuple, Union

from draccus import ChoiceRegistry, encode

from .testutils import *


class Color(Enum):
    blue = auto()
    red = auto()


def test_encode_basic_types():
    # Test basic types
    assert encode(42) == 42
    assert encode(3.14) == 3.14
    assert encode("hello") == "hello"
    assert encode(True) is True
    assert encode(None) is None


def test_encode_enum():
    assert encode(Color.blue) == "blue"
    assert encode(Color.red) == "red"


def test_encode_list():
    # Test list with and without type parameters
    assert encode([1, 2, 3]) == [1, 2, 3]
    assert encode(["a", "b", "c"]) == ["a", "b", "c"]
    assert encode([Color.blue, Color.red]) == ["blue", "red"]


def test_encode_tuple():
    # Test tuple with and without type parameters
    assert encode((1, 2, 3)) == [1, 2, 3]
    assert encode(("a", "b", "c")) == ["a", "b", "c"]
    assert encode((Color.blue, Color.red)) == ["blue", "red"]


def test_encode_dict():
    # Test dict with and without type parameters
    assert encode({"a": 1, "b": 2}) == {"a": 1, "b": 2}
    assert encode({"a": Color.blue, "b": Color.red}) == {"a": "blue", "b": "red"}
    assert encode({1: "a", 2: "b"}) == {1: "a", 2: "b"}


def test_encode_dataclass():
    @dataclass
    class SimpleClass:
        x: int
        y: str
        z: Optional[Color] = None

    obj = SimpleClass(42, "hello", Color.blue)
    expected = {"x": 42, "y": "hello", "z": "blue"}
    assert encode(obj) == expected

    obj = SimpleClass(42, "hello")
    expected = {"x": 42, "y": "hello", "z": None}
    assert encode(obj) == expected


def test_encode_nested_dataclass():
    @dataclass
    class Inner:
        a: int
        b: str

    @dataclass
    class Outer:
        inner: Inner
        c: List[Inner]

    obj = Outer(Inner(1, "a"), [Inner(2, "b"), Inner(3, "c")])
    expected = {"inner": {"a": 1, "b": "a"}, "c": [{"a": 2, "b": "b"}, {"a": 3, "b": "c"}]}
    assert encode(obj) == expected


def test_encode_generic_dataclass():
    @dataclass
    class Container(Generic[T]):
        value: T
        items: List[T]

    obj = Container[int](value=42, items=[1, 2, 3])
    expected = {"value": 42, "items": [1, 2, 3]}
    assert encode(obj) == expected

    obj = Container[str](value="hello", items=["a", "b"])
    expected = {"value": "hello", "items": ["a", "b"]}
    assert encode(obj) == expected


def test_encode_complex_nesting():
    @dataclass
    class Complicated:
        x: List[List[List[Dict[int, Tuple[int, float, str, List[float]]]]]]

    obj = Complicated([[[{0: (2, 1.23, "bob", [1.2, 1.3])}]]])
    expected = {"x": [[[{0: [2, 1.23, "bob", [1.2, 1.3]]}]]]}
    assert encode(obj) == expected


class Animal(ChoiceRegistry):
    pass


@Animal.register_subclass("dog")
@dataclass
class Dog(Animal):
    name: str
    age: int


@Animal.register_subclass("cat")
@dataclass
class Cat(Animal):
    name: str
    lives: int


@dataclass
class Zoo:
    animals: List[Animal]

    def __post_init__(self):
        assert all(isinstance(a, Animal) for a in self.animals)


def test_encode_choice_type():

    dog = Dog("Fido", 3)
    expected = {"type": "dog", "name": "Fido", "age": 3}
    assert encode(dog, Animal) == expected

    assert encode(dog) == {"name": "Fido", "age": 3}

    cat = Cat("Whiskers", 9)
    expected = {"type": "cat", "name": "Whiskers", "lives": 9}
    assert encode(cat, Animal) == expected
    expected = {"name": "Whiskers", "lives": 9}
    assert encode(cat) == expected


def test_encode_choice_type_in_generic():
    from draccus.choice_types import ChoiceRegistry

    dog = Dog("Fido", 3)
    cat = Cat("Whiskers", 9)

    zoo = Zoo([dog, cat])
    expected = {"animals": [{"type": "dog", "name": "Fido", "age": 3}, {"type": "cat", "name": "Whiskers", "lives": 9}]}
    assert encode(zoo) == expected

    @dataclass
    class Home:
        animals: List[Union[Dog, Cat]]

        def __post_init__(self):
            assert all(isinstance(a, Animal) for a in self.animals)

    home = Home([dog, cat])
    expected = {"animals": [{"name": "Fido", "age": 3}, {"name": "Whiskers", "lives": 9}]}
    assert encode(home) == expected


def test_encode_generic_union_member():
    @dataclass
    class Foo:
        x: Union[List[int], str]

        def __post_init__(self):
            assert isinstance(self.x, list) or isinstance(self.x, str)

    assert encode(Foo(x=[1, 2, 3])) == {"x": [1, 2, 3]}
    assert encode(Foo(x="hello")) == {"x": "hello"}


def test_encode_literal_union_member():
    @dataclass
    class Foo:
        x: Union[Literal["a", "b"], int]

        def __post_init__(self):
            assert isinstance(self.x, int) or self.x in ["a", "b"]

    assert encode(Foo(x=1)) == {"x": 1}
    assert encode(Foo(x="a")) == {"x": "a"}



import logging
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, ClassVar, List, Optional, Tuple, Type

import pytest

from .testutils import TestSetup

# List of simple attributes to use in tests:
simple_arguments: List[Tuple[Type, Any, Any]] = [
    # type, passed value, expected (parsed) value
    (int, "123", 123),
    (int, 123, 123),
    (int, "-1", -1),
    (float, "123.0", 123.0),
    (float, "'0.0'", 0.0),
    (float, "0.123", 0.123),
    (float, "0.123", 0.123),
    (float, 0.123, 0.123),
    (float, 123, 123.0),
    (bool, "True", True),
    (bool, "False", False),
    (bool, "true", True),
    (bool, "false", False),
    (str, "bob", "bob"),
    (str, "'bob'", "bob"),
    (str, "[123]", "[123]"),
    (str, "123", "123"),
]


@pytest.fixture(params=simple_arguments)
def simple_attribute(request):
    """Test fixture that produces an tuple of (type, passed value, expected value)"""
    some_type, passed_value, expected_value = request.param
    logging.debug(f"Attribute type: {some_type}, passed value: '{passed_value}', expected: '{expected_value}'")
    return request.param


@pytest.fixture
def assert_equals_stdout(capsys):
    def strip(string):
        return "".join(string.split())

    def should_equal(expected: str, file_path: Optional[str] = None):
        out = capsys.readouterr().out
        assert strip(out) == strip(expected), file_path

    return should_equal


@pytest.fixture
def no_stdout(capsys, caplog):
    """Asserts that no output was produced on stdout.

    Args:
        capsys (pytest.fixture): The capsys fixture
    """
    with caplog.at_level(logging.DEBUG):
        yield
    captured = capsys.readouterr()
    if captured.out != "":
        pytest.fail(f"Test generated some output in stdout: '{captured.out}'")
    if captured.err != "":
        pytest.fail(f"Test generated some output in stderr: '{captured.err}'")


@pytest.fixture
def no_warnings(caplog):
    yield
    for when in ("setup", "call"):
        messages = [x.message for x in caplog.get_records(when) if x.levelno == logging.WARNING]
        if messages:
            pytest.fail("warning messages encountered during testing: {}".format(messages))


@pytest.fixture
def silent(no_stdout, no_warnings):
    """
    Test fixture that will make a tests fail if it prints anything to stdout or
    logs warnings
    """
    pass


@pytest.fixture
def logs_warning(caplog):
    yield
    messages = [x.message for x in caplog.get_records("call") if x.levelno == logging.WARNING]
    if not messages:
        pytest.fail(f"No warning messages were logged: {messages}")


@pytest.fixture
def TaskHyperParameters():
    """Test fixture that gives a good example use-case from a real datascience
    project.
    """

    from enum import Enum

    class Activations(Enum):
        TANH = "TANH"
        RELU = "RELU"
        LINEAR = "LINEAR"

    @dataclass(frozen=True)
    class TaskHyperParameters(TestSetup):
        """
        HyperParameters for a task-specific model
        """

        name: str  # name of the task
        num_layers: int = 1  # number of dense layers
        num_units: int = 8  # units per layer
        activation: Activations = field(default=Activations.TANH)  # activation function
        use_batchnorm: bool = False  # whether or not to use batch normalization after each dense layer
        use_dropout: bool = True  # whether or not to use dropout after each dense layer
        dropout_rate: float = 0.1  # the dropout rate
        use_image_features: bool = True  # whether or not image features should be used as input
        use_likes: bool = True  # whether or not 'likes' features should be used as input
        l1_reg: float = 0.005  # L1 regularization coefficient
        l2_reg: float = 0.005  # L2 regularization coefficient

        # Whether or not a task-specific Embedding layer should be used on the 'likes' features.
        # When set to 'True', it is expected that there no shared embedding is used.
        embed_likes: bool = False

    return TaskHyperParameters


class Optimizers(Enum):
    ADAM = "ADAM"
    SGD = "SGD"


@pytest.fixture
def HyperParameters(TaskHyperParameters):
    @dataclass(frozen=True)
    class HyperParameters(TestSetup):
        """Hyperparameters of a multi-headed model."""

        batch_size: int = 128  # the batch size
        learning_rate: float = 0.001  # Learning Rate
        optimizer: Optimizers = Optimizers.SGD  # Which optimizer to use during training.

        # number of individual 'pages' that were kept during preprocessing of the 'likes'.
        # This corresponds to the number of entries in the multi-hot like vector.
        num_like_pages: int = 10_000

        gender_loss_weight: float = 1.0  # relative weight of the gender loss
        age_loss_weight: float = 1.0  # relative weight of the age_group loss

        num_text_features: ClassVar[int] = 91
        num_image_features: ClassVar[int] = 65

        max_number_of_likes: int = 2000
        embedding_dim: int = 8

        shared_likes_embedding: bool = True

        # Whether or not to better filtering of liked pages
        use_custom_likes: bool = True

        # Gender model settings
        gender: TaskHyperParameters = TaskHyperParameters(
            "gender",
            num_layers=1,
            num_units=32,
            use_batchnorm=False,
            use_dropout=True,
            dropout_rate=0.1,
            use_image_features=True,
            use_likes=True,
        )

        # Age Group Model settings
        age_group: TaskHyperParameters = TaskHyperParameters(
            "age_group",
            num_layers=2,
            num_units=64,
            use_batchnorm=False,
            use_dropout=True,
            dropout_rate=0.1,
            use_image_features=True,
            use_likes=True,
        )

        # Personality Model(s) settings:
        personality: TaskHyperParameters = TaskHyperParameters(
            "personality",
            num_layers=1,
            num_units=8,
            use_batchnorm=False,
            use_dropout=True,
            dropout_rate=0.1,
            use_image_features=False,
            use_likes=False,
        )

    return HyperParameters



__version__ = "0.8.0"

from .argparsing import parse, wrap
from .cfgparsing import dump, load, loads
from .choice_types import CHOICE_TYPE_KEY, ChoiceRegistry, ChoiceType, PluginRegistry
from .fields import field
from .options import ConfigType, Options, config_type
from .parsers.decoding import decode
from .parsers.encoding import encode
from .utils import ParsingError

get_config_type = Options.get_config_type
set_config_type = Options.set_config_type

__all__ = [
    "CHOICE_TYPE_KEY",
    "ChoiceRegistry",
    "ChoiceType",
    "ConfigType",
    "Options",
    "ParsingError",
    "PluginRegistry",
    "config_type",
    "decode",
    "dump",
    "encode",
    "field",
    "get_config_type",
    "load",
    "parse",
    "set_config_type",
    "wrap",
]



"""Tests for literal type support in Draccus."""
import io
import sys
from contextlib import redirect_stdout
from dataclasses import dataclass, field
from typing import Literal

import pytest

from draccus import ParsingError, decode, parse

from .testutils import TestSetup, raises


@dataclass
class StringLiteralConfig(TestSetup):
    mode: Literal["train", "test", "eval"] = field(default="train")
    size: Literal["small", "medium", "large"] = field(default="small")


@dataclass
class NumericLiteralConfig(TestSetup):
    mode: Literal[1, 2, 3] = field(default=1)
    size: Literal[10, 20, 30] = field(default=10)


@dataclass
class MixedLiteralConfig(TestSetup):
    mode: Literal["train", 1, True] = field(default="train")
    size: Literal["small", 20, False] = field(default="small")


def test_string_literal_valid():
    config = decode(StringLiteralConfig, {"mode": "train"})
    assert config.mode == "train"

    config = decode(StringLiteralConfig, {"mode": "eval"})
    assert config.mode == "eval"

    config = decode(StringLiteralConfig, {"mode": "test"})
    assert config.mode == "test"


def test_string_literal_invalid():
    with pytest.raises(Exception) as exc_info:
        decode(StringLiteralConfig, {"mode": "invalid"})
    assert "into one of" in str(exc_info.value)


def test_numeric_literal_valid():
    config = decode(NumericLiteralConfig, {"size": 10})
    assert config.size == 10

    config = decode(NumericLiteralConfig, {"size": 20})
    assert config.size == 20

    config = decode(NumericLiteralConfig, {"size": 30})
    assert config.size == 30


def test_numeric_literal_invalid():
    with pytest.raises(Exception) as exc_info:
        decode(NumericLiteralConfig, {"size": 15})
    assert "into one of" in str(exc_info.value)


def test_mixed_literal_valid():
    # Test string value
    config = decode(MixedLiteralConfig, {"mode": "train", "size": "small"})
    assert config.mode == "train"
    assert config.size == "small"

    # Test numeric value
    config = decode(MixedLiteralConfig, {"mode": 1, "size": 20})
    assert config.mode == 1
    assert config.size == 20

    # Test boolean value
    config = decode(MixedLiteralConfig, {"mode": True, "size": False})
    assert config.mode is True
    assert config.size is False


def test_mixed_literal_invalid():
    with pytest.raises(Exception) as exc_info:
        decode(MixedLiteralConfig, {"mode": "invalid", "size": "small"})
    assert "into one of" in str(exc_info.value)

    with pytest.raises(Exception) as exc_info:
        decode(MixedLiteralConfig, {"mode": "train", "size": 15})
    assert "into one of" in str(exc_info.value)


def test_literal_help_text():
    """Test that help text is generated correctly for literal types."""
    # Test string literal help
    config = StringLiteralConfig.get_help_text("--help")
    help_text = str(config)
    assert "mode" in help_text
    assert "train" in help_text
    assert "eval" in help_text
    assert "test" in help_text
    assert "{small,medium,large}" in help_text

    # Test numeric literal help
    config = NumericLiteralConfig.get_help_text("--help")
    help_text = str(config)
    assert "size" in help_text
    assert "10" in help_text
    assert "20" in help_text
    assert "30" in help_text
    assert "{10,20,30}" in help_text

    # Test mixed literal help
    config = MixedLiteralConfig.get_help_text("--help")
    help_text = str(config)
    assert "mode" in help_text
    assert "train" in help_text
    assert "1" in help_text
    assert "True" in help_text
    assert "size" in help_text
    assert "small" in help_text
    assert "20" in help_text
    assert "False" in help_text


def test_literal_argparse_valid():
    """Test that literal types work correctly with argparse."""
    # Test string literal
    config = parse(StringLiteralConfig, args=["--mode", "train"])
    assert config.mode == "train"

    config = parse(StringLiteralConfig, args=["--mode", "eval"])
    assert config.mode == "eval"

    # Test numeric literal
    config = parse(NumericLiteralConfig, args=["--size", "10"], exit_on_error=False)
    assert config.size == 10

    config = parse(NumericLiteralConfig, args=["--size", "30"])
    assert config.size == 30

    # Test mixed literals
    config = parse(MixedLiteralConfig, args=["--mode", "train", "--size", "small"])
    assert config.mode == "train"
    assert config.size == "small"

    config = parse(MixedLiteralConfig, args=["--mode", "1", "--size", "20"])
    assert config.mode == 1
    assert config.size == 20

    config = parse(MixedLiteralConfig, args=["--mode", "True", "--size", "False"])
    assert config.mode is True
    assert config.size is False


def test_literal_argparse_invalid():
    """Test that invalid literal values are caught by argparse."""
    # Test string literal
    with pytest.raises(SystemExit):
        parse(StringLiteralConfig, args=["--mode", "invalid"])

    # Test numeric literal
    with pytest.raises(SystemExit):
        parse(NumericLiteralConfig, args=["--size", "15"])

    # Test mixed literals
    with pytest.raises(SystemExit):
        parse(MixedLiteralConfig, args=["--mode", "invalid", "--size", "20"])

    with pytest.raises(SystemExit):
        parse(MixedLiteralConfig, args=["--mode", "train", "--size", "3"])



"""Utility functions used in various parts of the draccus package."""
import builtins
import collections.abc as c_abc
import dataclasses
import enum
import inspect
import logging
import sys
import types
from dataclasses import _MISSING_TYPE
from enum import Enum
from logging import getLogger
from typing import (
    Any,
    ClassVar,
    Container,
    Dict,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Protocol,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import typing_inspect as tpi

try:
    from typing import get_args
except ImportError:

    def get_args(some_type: Type) -> Tuple[Type, ...]:  # type: ignore
        return getattr(some_type, "__args__", ())


# try:
#     from typing import get_origin  # type: ignore
# except ImportError:
#     pass  # type: ignore


try:
    from _typeshed import DataclassInstance  # type: ignore
except ImportError:

    class DataclassInstance(Protocol):  # type: ignore
        __dataclass_fields__: ClassVar[Dict[str, dataclasses.Field]]


logger = getLogger(__name__)

builtin_types = [getattr(builtins, d) for d in dir(builtins) if isinstance(getattr(builtins, d), type)]

T = TypeVar("T")

Dataclass = TypeVar("Dataclass", bound=DataclassInstance)
DataclassType = Type[Dataclass]


class DraccusException(Exception):
    pass


class ParsingError(DraccusException):
    pass


class DecodingError(ParsingError):
    key_path: Sequence[str]
    message: str

    def strip_prefix(self, prefix: Sequence[str]) -> "DecodingError":
        if len(self.key_path) < len(prefix):
            return self
        if not all(self.key_path[i] == prefix[i] for i in range(len(prefix))):
            raise ValueError(f"Prefix {prefix} is not a prefix of {self.key_path}")
        return DecodingError(self.key_path[len(prefix) :], self.message)

    def __init__(self, key_path: Sequence[str], message: str):
        self.key_path = key_path
        self.message = message
        super().__init__(key_path, message)

    def __str__(self):
        if len(self.key_path) == 0:
            return self.message

        key_path = ".".join(self.key_path)

        return f"`{key_path}`: {self.message}"


def get_item_type(container_type: Type[Container[T]]) -> Type[T]:
    """Returns the `type` of the items in the provided container `type`.

    When no type annotation is found, or no item type is found, returns
    `typing.Any`.
    NOTE: If a type with multiple arguments is passed, only the first type
    argument is returned.
    """
    if container_type in {
        list,
        set,
        tuple,
        List,
        Set,
        Tuple,
        Dict,
        Mapping,
        MutableMapping,
    }:
        # TODO(dlwh): I don't like this
        # the built-in `list` and `tuple` types don't have annotations for their item types.
        return Any  # type: ignore
    type_arguments = getattr(container_type, "__args__", None)
    if type_arguments:
        return type_arguments[0]
    else:
        return Any  # type: ignore


def _mro(t: Type) -> Sequence[Type]:
    # TODO: This is mostly used in 'is_tuple' and such, and should be replaced with
    # either the built-in 'get_origin' from typing, or from typing-inspect.
    if t is None:
        return []
    if hasattr(t, "__mro__"):
        return t.__mro__
    elif tpi.get_origin(t) is type:
        return []
    elif hasattr(t, "mro") and callable(t.mro):
        return t.mro()
    return []


def is_list(t: Type) -> bool:
    return list in _mro(t)


def is_tuple(t: Type) -> bool:
    return tuple in _mro(t)


def is_dict(t: Type) -> bool:
    mro = _mro(t)
    return dict in mro or Mapping in mro or c_abc.Mapping in mro


def is_set(t: Type) -> bool:
    mro = _mro(t)
    return set in mro


def is_dataclass_type(t: Type) -> bool:
    """Returns whether t is a dataclass type or a TypeVar of a dataclass type.

    Args:
        t (Type): Some type.

    Returns:
        bool: Whether its a dataclass type.
    """
    return dataclasses.is_dataclass(t) or (tpi.is_typevar(t) and dataclasses.is_dataclass(tpi.get_bound(t)))


def is_enum(t: Type) -> bool:
    if inspect.isclass(t):
        return issubclass(t, enum.Enum)
    return Enum in _mro(t)


def is_bool(t: Type) -> bool:
    return bool in _mro(t)


def is_tuple_or_list(t: Type) -> bool:
    return is_list(t) or is_tuple(t)


def is_union(t: Type) -> bool:
    if getattr(t, "__origin__", "") == Union:
        return True

    elif sys.version_info >= (3, 9):
        return tpi.is_union_type(t)

    return False


def is_literal(t: Type) -> bool:
    """Returns True if t is a Literal type."""
    try:
        from typing import Literal

        return getattr(t, "__origin__", None) is Literal
    except ImportError:
        # For Python < 3.8
        return False


def is_homogeneous_tuple_type(t: Type[Tuple]) -> bool:
    if not is_tuple(t):
        return False
    type_arguments = get_type_arguments(t)
    if not type_arguments:
        return True
    assert isinstance(type_arguments, tuple), type_arguments
    if len(type_arguments) == 2 and type_arguments[1] is Ellipsis:
        return True
    # TODO: Not sure if this will work with more complex item times (like nested tuples)
    return len(set(type_arguments)) == 1


def is_optional(t: Type) -> bool:
    return is_union(t) and type(None) in get_type_arguments(t)


def is_tuple_or_list_of_dataclasses(t: Type) -> bool:
    return is_tuple_or_list(t) and is_dataclass_type(get_item_type(t))


def contains_dataclass_type_arg(t: Type) -> bool:
    if is_dataclass_type(t):
        return True
    elif is_tuple_or_list_of_dataclasses(t):
        return True
    elif is_union(t):
        return any(contains_dataclass_type_arg(arg) for arg in get_type_arguments(t))
    return False


def get_dataclass_type_arg(t: Type) -> Optional[Type]:
    if not contains_dataclass_type_arg(t):
        return None
    if is_dataclass_type(t):
        return t
    elif is_tuple_or_list(t) or is_union(t):
        return next(
            filter(None, (get_dataclass_type_arg(arg) for arg in get_type_arguments(t))),
            None,
        )
    return None


def is_optional_or_union_with_dataclass_type_arg(t: Type) -> bool:
    if is_union(t) or is_optional(t):
        return any(
            is_dataclass_type(arg) or is_optional_or_union_with_dataclass_type_arg(arg) for arg in get_type_arguments(t)
        )
    return False


def get_type_arguments(container_type: Type) -> Tuple[Type, ...]:
    # return getattr(container_type, "__args__", ())
    return get_args(container_type)


def get_type_name(some_type: Type):
    result = getattr(some_type, "__name__", str(some_type))
    type_arguments = get_type_arguments(some_type)
    if type_arguments:
        result += f"[{','.join(get_type_name(T) for T in type_arguments)}]"
    return result


def default_value(field: dataclasses.Field) -> Union[T, _MISSING_TYPE]:
    """Returns the default value of a field in a dataclass, if available.
    When not available, returns `dataclasses.MISSING`.

    Args:
        field (dataclasses.Field): The dataclasses.Field to get the default value of.

    Returns:
        Union[T, _MISSING_TYPE]: The default value for that field, if present, or None otherwise.
    """
    if field.default is not dataclasses.MISSING:
        return field.default
    elif field.default_factory is not dataclasses.MISSING:  # type: ignore
        constructor = field.default_factory  # type: ignore
        return constructor()
    else:
        return dataclasses.MISSING


def get_defaults_dict(c: Dataclass):
    """ " Get defaults of a dataclass without generating the object"""
    defaults_dict = {}
    for field in dataclasses.fields(c):  # type: ignore
        if is_dataclass_type(field.type):
            defaults_dict[field.name] = get_defaults_dict(cast(DataclassType, field.type))  # type: ignore
        else:
            if field.default is not dataclasses.MISSING:
                defaults_dict[field.name] = field.default
            elif field.default_factory is not dataclasses.MISSING:
                try:
                    defaults_dict[field.name] = field.default_factory()
                except Exception as e:
                    logging.debug(
                        f"Failed getting default for field {field.name} using its default factory.\n\tUnderlying"
                        f" error: {e}"
                    )
                    continue
    return defaults_dict


def keep_keys(d: Dict, keys_to_keep: Iterable[str]) -> Tuple[Dict, Dict]:
    d_keys = set(d.keys())  # save a copy since we will modify the dict.
    removed = {}
    for key in d_keys:
        if key not in keys_to_keep:
            removed[key] = d.pop(key)
    return d, removed


def flatten(d, parent_key=None, sep="."):
    items = []
    for k, v in d.items():
        new_key = f"{parent_key}{sep}{k}" if parent_key else k
        if isinstance(v, c_abc.MutableMapping):
            items.extend(flatten(v, parent_key=new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def deflatten(d: Dict[str, Any], sep: str = "."):
    deflat_d: Dict[str, Any] = {}
    for key in d:
        key_parts = key.split(sep)
        curr_d = deflat_d
        for inner_key in key_parts[:-1]:
            if inner_key not in curr_d:
                curr_d[inner_key] = {}
            curr_d = curr_d[inner_key]
        curr_d[key_parts[-1]] = d[key]
    return deflat_d


def remove_matching(dict_a, dict_b):
    dict_a = flatten(dict_a)
    dict_b = flatten(dict_b)
    for key in dict_b:
        if key in dict_a and dict_b[key] == dict_a[key]:
            del dict_a[key]
    return deflatten(dict_a)


def format_error(e: Exception):
    try:
        return f"{type(e).__name__}: {e}"
    except Exception:
        return f"Exception: {e}"


def is_generic_arg(arg):
    try:
        return arg.__name__ in ["KT", "VT", "T"]
    except Exception:
        return False


def has_generic_arg(args):
    for arg in args:
        if is_generic_arg(arg):
            return True
    return False


def is_choice_type(cls: Any) -> bool:
    """
    Returns True if
    1) cls is a ChoiceRegistry or PluginRegistry, or a *direct* subclass of one of those
    OR
    2) cls structurally matches the ChoiceType protocol, but none of its parents do
    """
    from draccus.choice_types import ChoiceRegistry, ChoiceType, PluginRegistry

    CHOICE_BASES = (ChoiceRegistry, PluginRegistry, ChoiceType)

    # Skip if not a proper class
    if not isinstance(cls, type):
        return False

    # 1. does not structurally match protocol --> False
    try:
        if not issubclass(cls, ChoiceType):
            return False
    except TypeError:
        return False  # Generic alias like list[int]

    # 2. Direct subclass of known base registry --> True
    if any(base in cls.__bases__ for base in CHOICE_BASES):
        return True

    # 3. none of its parents match structurally --> True
    for base in cls.__bases__:
        try:
            if issubclass(base, ChoiceType):
                return False
        except TypeError:
            continue

    return True


class StringHolderEnum(type):
    """Like a python enum but just holds string constants, as opposed to wrapped string constants"""

    # https://stackoverflow.com/questions/62881486/a-group-of-constants-in-python

    def __new__(cls, name, bases, members):
        # this just iterates through the class dict and removes
        # all the dunder methods
        cls.members = [v for k, v in members.items() if not k.startswith("__") and not callable(v)]
        return super().__new__(cls, name, bases, members)

    # giving your class an __iter__ method gives you membership checking
    # and the ability to easily convert to another iterable
    def __iter__(cls):
        yield from cls.members


def canonicalize_union(t: Type):
    if sys.version_info >= (3, 10) and isinstance(t, types.UnionType):
        return Union[tuple(canonicalize_union(u) for u in t.__args__)]
    # recursively canonicalize
    args = get_args(t)
    if args:
        args = tuple(canonicalize_union(arg) for arg in args)
        origin = tpi.get_origin(t)
        if origin is not None:
            origin = raw_to_typing.get(origin, origin)
            return origin[args]
        else:
            return Union[args]

    return t


CONFIG_ARG = "config_path"

raw_to_typing = {
    dict: Dict,
    list: List,
    tuple: Tuple,
    set: Set,
}


def stringify_type(tpe: Type):
    origin = tpi.get_origin(tpe)
    if tpe is float:
        return "float"
    elif tpe is int:
        return "int"
    elif tpe is bool:
        return "bool"
    elif tpe is str:
        return "str"
    elif tpe is list:
        return "list"
    elif origin is list or origin is List:
        return f"list[{stringify_type(get_args(tpe)[0])}]"

    # TODO: more types
    else:
        # strip any package names/get plain name
        try:
            return tpe.__name__
        except Exception:
            return str(tpe)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import os
from pathlib import Path
from typing import Optional, TextIO, Type, Union

from draccus import utils
from draccus.options import Options, config_type
from draccus.parsers.decoding import decode
from draccus.parsers.encoding import encode
from draccus.utils import Dataclass


def parse_string(s: str) -> dict:
    """
    Parse a string into a dictionary using the current config type parser.

    Args:
        s: The string to parse

    Returns:
        A dictionary containing the parsed configuration
    """
    parser = Options.get_config_type().value
    return parser.parse_string(s)


def load_config(
    stream: Union[str, TextIO, os.PathLike], *, file: Optional[Union[str, Path, os.PathLike]] = None
) -> dict:
    """
    Load configuration from a stream (file object or string) or file path.

    Args:
        stream: Either a file object, string content, or file path
        file: Optional file path used to determine the config type based on extension

    Returns:
        A dictionary containing the loaded configuration

    Note:
        If file is provided, the config type will be determined by the file extension.
        Supported extensions: .toml, .json, .yaml, .yml
    """
    if file is not None:
        fpath = str(file)
        if fpath.endswith(".toml"):
            with config_type("toml"):
                return load_config(stream)
        elif fpath.endswith(".json"):
            with config_type("json"):
                return load_config(stream)
        elif fpath.endswith(".yaml") or fpath.endswith(".yml"):
            with config_type("yaml"):
                return load_config(stream)

    parser = Options.get_config_type().value
    try:
        return parser.load_config(stream)
    except Exception as e:  # pylint: disable=broad-except
        raise utils.ParsingError(f"Failed to load config from {stream}") from e


def save_config(d: dict, stream=None, **kwargs) -> Optional[str]:
    """
    Save a configuration dictionary to a stream or return as a string.

    Args:
        d: The configuration dictionary to save
        stream: Optional stream to write to. If None, returns the configuration as a string
        **kwargs: Additional arguments passed to the parser's save_config method

    Returns:
        If stream is None, returns the configuration as a string.
        Otherwise, returns None after writing to the stream.
    """
    parser = Options.get_config_type().value
    return parser.save_config(d, stream, **kwargs)


def load(t: Type[Dataclass], stream: Union[str, TextIO, os.PathLike]) -> Dataclass:
    """
    Load a config from a file path, file object, or string.

    Args:
        t: The dataclass type to load into
        stream: Either a file path, file object, or string content

    Returns:
        An instance of the specified dataclass with values loaded from the stream

    Note:
        For string content, consider using loads() instead for clarity.
        This method maintains backwards compatibility with previous versions.
    """
    if isinstance(stream, (str, os.PathLike)) and os.path.exists(stream):
        # If stream is a file path, open it
        with open(stream, "r") as f:
            dictionary = load_config(f, file=stream)
    else:
        # If stream is a file object or string content
        dictionary = load_config(stream)

    return decode(t, dictionary)


def loads(t: Type[Dataclass], s: str) -> Dataclass:
    """
    Load a config from a string.

    Args:
        t: The dataclass type to load into
        s: The string containing the configuration

    Returns:
        An instance of the specified dataclass with values loaded from the string
    """
    dictionary = load_config(s)
    return decode(t, dictionary)


def dump(config: Dataclass, stream=None, omit_defaults: bool = False, **kwargs) -> Optional[str]:
    """
    Dump the config object to a stream or return as a string.

    Args:
        config: The dataclass instance to dump
        stream: Optional stream to write to. If None, returns the configuration as a string
        omit_defaults: If True, omits any values that match their default values
        **kwargs: Additional arguments passed to the parser's save_config method

    Returns:
        If stream is None, returns the configuration as a string.
        Otherwise, returns None after writing to the stream.
    """
    config_dict = encode(config)
    if omit_defaults:
        defaults_dict = encode(utils.get_defaults_dict(config))
        config_dict = utils.remove_matching(config_dict, defaults_dict)
    return save_config(config_dict, stream, **kwargs)



"""Simple, extendable mechanism for encoding pracitaclly anything to string.

Just register a new encoder for a given type like so:

import draccus
import numpy as np
@draccus.encode.register
def encode_ndarray(obj: np.ndarray) -> str:
    return obj.tostring()
"""
import json
import typing
from argparse import Namespace
from collections.abc import Mapping
from dataclasses import fields, is_dataclass
from enum import Enum
from logging import getLogger
from os import PathLike
from typing import Any, Dict, Hashable, List, Optional, Tuple, Type, Union

from draccus.choice_types import CHOICE_TYPE_KEY
from draccus.parsers.registry_utils import RegistryFunc, withregistry
from draccus.utils import is_choice_type

logger = getLogger(__name__)


class SimpleJsonEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        return encode(o)


"""
# NOTE: This code is commented because of static typing check error.
# The problem is incompatibility of mypy and singledispatch.
# See mypy issues for more info:
# https://github.com/python/mypy/issues/8356
# https://github.com/python/mypy/issues/2904
# https://github.com/python/mypy/issues/9112#issuecomment-725316936

class Dataclass(Protocol):
    # see dataclasses.is_dataclass implementation with _FIELDS
    __dataclass_fields__: Dict[str, Field[Any]]


T = TypeVar("T", bool, int, None, str)


@overload
def encode(obj: Dataclass) -> Dict[str, Any]: ...

@overload
def encode(obj: Union[List[Any], Set[Any], Tuple[Any, ...]]) -> List[Any]:
    ...

@overload
def encode(obj: Mapping[Any, Any]) -> Dict[Any, Any]: ...

@overload
def encode(obj: T) -> T: ...
"""


@withregistry
def encode(obj: Any, declared_type: Optional[Type] = None) -> Any:
    """Encodes an object into a json/yaml-compatible primitive type.

    Args:
        obj: The object to encode
        declared_type: Optional type annotation for the object. Used to determine if the object
                      should be encoded as a choice type based on its declared type rather than
                      its concrete type.
    """
    if declared_type is not None:
        underlying_type = typing.get_origin(declared_type) or declared_type
        # we have to handle unions specially for declared types:
        if underlying_type is Union:
            # find the first type that matches the object's type
            for t in typing.get_args(declared_type):
                # we can't use subscripted generic types here
                if typing.get_origin(t) is typing.Literal:
                    for arg in typing.get_args(t):
                        if arg == obj:
                            underlying_type = t
                            declared_type = t
                            break
                elif isinstance(obj, typing.get_origin(t) or t):
                    underlying_type = typing.get_origin(t) or t
                    declared_type = t
                    break
    else:
        underlying_type = type(obj)
    cached_func: RegistryFunc = encode.dispatch(underlying_type)

    if cached_func is None:
        # see if the actual type has a custom encoder
        cached_func = encode.dispatch(type(obj))

    if cached_func is not None:
        fn = cached_func.func

        # we want to support the old interface where the decoding function
        # takes only one argument, so we wrap it here
        try:
            return fn(obj, declared_type)
        except TypeError:
            try:
                return fn(obj)
            except Exception as e:  # pylint: disable=broad-except
                raise Exception(f"Couldn't encode {obj}") from e

    try:
        if underlying_type is not None and is_choice_type(underlying_type):
            return encode_choice(obj, underlying_type)
        elif is_dataclass(obj):
            return encode_dataclass(obj, declared_type)
        elif obj is None:
            return None
        else:
            raise Exception(f"No parser for object {obj} of type {type(obj)}, consider using draccus.encode.register")
    except Exception as e:
        logger.debug(f"Cannot encode object {obj}: {e}")
        raise e


def encode_dataclass(obj: Any, declared_type: Optional[Type] = None):
    d: Dict[str, Any] = dict()

    # Handle type parameters if declared_type is provided
    type_map = {}
    if declared_type is not None:
        origin = typing.get_origin(declared_type)
        if origin is not None and hasattr(origin, "__parameters__"):
            type_args = typing.get_args(declared_type)
            type_vars = origin.__parameters__
            type_map = dict(zip(type_vars, type_args))
        else:
            origin = declared_type

    for field in fields(obj):
        value = getattr(obj, field.name)
        try:
            # If we have a type map, use it to resolve the field's type
            field_type = field.type
            if type_map and hasattr(field_type, "__parameters__"):
                field_type = field_type[tuple(type_map.get(p, p) for p in field_type.__parameters__)]
            d[field.name] = encode(value, field_type)
        except TypeError as e:
            logger.error(f"Unable to encode field {field.name}: {e}")
            raise e
    return d


def encode_choice(obj: Any, declared_type: Type) -> Dict[str, Any]:
    """Encodes an object as a choice type based on its declared type.

    Args:
        obj: The object to encode
        declared_type: The type annotation for the object, which must be a choice type
    """
    if not is_choice_type(declared_type):
        raise ValueError(f"Expected a choice type, got {declared_type}")

    encoded = encode_dataclass(obj, declared_type)

    if not isinstance(encoded, dict):
        raise Exception(f"Choice Class {obj} is not encoded as a dict: {encoded}")

    # Get the choice name from the declared type
    choice_name = obj.get_choice_name(type(obj))
    encoded = {CHOICE_TYPE_KEY: choice_name, **encoded}

    return encoded


@encode.register(Mapping, include_subclasses=True)
def encode_dict(
    obj: Mapping, declared_type: Optional[Type] = None
) -> Union[typing.Mapping[Any, Any], List[Tuple[Any, Any]]]:
    constructor = type(obj)
    result: Union[Mapping, List[Tuple[Any, Any]]] = constructor()

    # Handle type parameters if declared_type is provided
    key_type = None
    value_type = None
    if declared_type is not None:
        type_args = typing.get_args(declared_type)
        if len(type_args) >= 2:
            key_type, value_type = type_args[:2]

    for k, v in obj.items():
        k_ = encode(k, key_type)
        v_ = encode(v, value_type)
        if isinstance(result, list):
            result.append((k_, v_))
        elif isinstance(k_, Hashable):
            result[k_] = v_  # type: ignore
        else:
            # If the encoded key isn't "Hashable", then we store it as a list of tuples
            if isinstance(result, Mapping):
                result = list(result.items())
            result.append((k_, v_))
    return result


@encode.register(Enum, include_subclasses=True)
def encode_enum(obj: Enum, declared_type: Optional[Type] = None) -> str:
    return obj.name


for t in [str, float, int, bool, bytes]:
    encode.register(t, lambda x, _=None: x)


@encode.register(list)
def encode_list(obj: list, declared_type: Optional[Type] = None) -> list:
    item_type = None
    if declared_type is not None:
        type_args = typing.get_args(declared_type)
        if type_args:
            item_type = type_args[0]
    return [encode(x, item_type) for x in obj]


@encode.register(tuple)
def encode_tuple(obj: tuple, declared_type: Optional[Type] = None) -> list:
    if declared_type is not None:
        type_args = typing.get_args(declared_type)
        if type_args and len(type_args) == len(obj):
            return [encode(x, t) for x, t in zip(obj, type_args)]
        elif type_args and len(type_args) == 2 and type_args[1] is Ellipsis:
            item_type = type_args[0]
            return [encode(x, item_type) for x in obj]
    return [encode(x) for x in obj]


@encode.register(set)
def encode_set(obj: set, declared_type: Optional[Type] = None) -> list:
    item_type = None
    if declared_type is not None:
        type_args = typing.get_args(declared_type)
        if type_args:
            item_type = type_args[0]
    return [encode(x, item_type) for x in obj]


encode.register(PathLike, lambda x, _=None: x.__fspath__())

encode.register(Namespace, lambda x, _=None: encode(vars(x)))



import dataclasses

import pytest

import draccus
from draccus import ParsingError
from draccus.choice_types import ChoiceRegistry
from draccus.utils import DecodingError


@dataclasses.dataclass
class Person(ChoiceRegistry):
    name: str  # Person's name


@dataclasses.dataclass
class Adult(Person):
    age: int


@dataclasses.dataclass
class Child(Person):
    favorite_toy: str


Person.register_subclass("adult", Adult)
Person.register_subclass("child", Child)


def test_choice_registry_decode():
    assert draccus.decode(Person, {"type": "adult", "name": "bob", "age": 10}) == Adult("bob", 10)
    assert draccus.decode(Person, {"type": "child", "name": "bob", "favorite_toy": "truck"}) == Child("bob", "truck")

    with pytest.raises(ParsingError):
        draccus.decode(Person, {"type": "baby", "name": "bob"})

    with pytest.raises(DecodingError):
        draccus.decode(Person, {"type": "adult", "name": "bob", "age": 10, "favorite_toy": "truck"})

    with pytest.raises(DecodingError):
        draccus.decode(Person, {"type": "adult", "name": 3})


def test_registry_decode_subtype_without_type():
    draccus.decode(Child, {"name": "bob", "favorite_toy": "truck"})

    with pytest.raises(DecodingError):
        draccus.decode(Child, {"type": "adult", "name": "bob", "age": 10})


def test_choice_registry_encode():
    assert draccus.encode(Adult("bob", 10), Person) == {"type": "adult", "name": "bob", "age": 10}
    assert draccus.encode(Child("bob", "truck"), Person) == {"type": "child", "name": "bob", "favorite_toy": "truck"}


def test_is_choicetype():
    assert draccus.utils.is_choice_type(Person)
    assert not draccus.utils.is_choice_type(Adult)
    assert not draccus.utils.is_choice_type(Child)



import tempfile
from enum import Enum, auto
from pathlib import Path

from tests.conftest import Optimizers


class Color(Enum):
    blue: str = auto()  # type: ignore
    red: str = auto()  # type: ignore


def test_load_hyperparameters_with_include(HyperParameters):
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)

        config_path = tmpdir / "config.yaml"
        config_path.write_text(
            """
age_group: !include ./age_group.yaml
batch_size: 1234
            """
        )

        age_group_path = tmpdir / "age_group.yaml"
        age_group_path.write_text(
            """
name: age_group
num_units: 16
"""
        )

        age_group_path_cli = tmpdir / "age_group_cli.yaml"
        age_group_path_cli.write_text(
            """
name: age_group_cli
num_units: 16
"""
        )


        config = HyperParameters.setup(f"--config_path {config_path} "
                                       f"--age_group 'include {age_group_path_cli}' "
                                       f"--age_group.num_units 32")

        assert config.age_group.name == "age_group_cli"
        assert config.age_group.num_units == 32


def test_merge_include_hyperparameters(HyperParameters):
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)

        config_path = tmpdir / "config.yaml"
        config_path.write_text(
            """
age_group: 
  name: age_group_1
  num_units: 16
batch_size: 1234
            """
        )

        age_group_path_cli = tmpdir / "age_group_cli.yaml"
        age_group_path_cli.write_text(
            """
name: age_group_cli
num_units: 678
"""
        )

        config = HyperParameters.setup(f"--config_path {config_path} "
                                       f"--age_group 'include {age_group_path_cli}' ")

        assert config.age_group.name == "age_group_cli"
        assert config.age_group.num_units == 678




from dataclasses import dataclass, field

import draccus

@dataclass
class TrainConfig:
    workers: int

@draccus.wrap()
def function_with_draccus_wrap(cfg: TrainConfig):
    assert cfg.workers == 8


def test_calling_using_dataclasses():
    config = TrainConfig(workers=8)
    function_with_draccus_wrap(config)


from argparse import ArgumentParser, _ArgumentGroup


class SuppressingArgumentParser(ArgumentParser):
    """
    ArgumentParser that has a slightly exotic method of handling conflicts.
    We don't want to raise an error if we have a conflict, we just want to ignore it.
    Ideally we'd have a way to say "take the first one", but that doesn't exist and
    can't be easily retrofitted. So we take the last one, but keep the
    old ones for the help message by leaving them in _group_actions but not in _actions.
    """

    def __init__(self, *args, **kwargs):
        kwargs = {**kwargs, "conflict_handler": "ignore"}
        super().__init__(*args, **kwargs)

    def add_argument_group(self, *args, **kwargs):
        group = _SuppressingArgumentGroup(self, *args, **kwargs)
        self._action_groups.append(group)
        return group

    def _handle_conflict_ignore(self, action, conflicting_actions):
        # remove all conflicting options
        for _option_string, action in conflicting_actions:
            action.container._remove_action(action)


class _SuppressingArgumentGroup(_ArgumentGroup):
    def __init__(self, container, *args, **kwargs):
        kwargs = {**kwargs, "conflict_handler": "ignore"}
        super().__init__(container, *args, **kwargs)

    def add_argument(self, *args, **kwargs):
        return super().add_argument(*args, **kwargs)

    def add_argument_group(self, *args, **kwargs):
        raise NotImplementedError("It's a bad idea to nest argument groups. argparse ignores them for help")

    def add_mutually_exclusive_group(self, *args, **kwargs):
        raise NotImplementedError("It's a bad idea to nest argument groups. argparse ignores them for help")

    def _remove_action(self, action):
        self._actions.remove(action)
        # don't remove from _group_actions, so that we can still show the old ones in the help message

    def _handle_conflict_ignore(self, action, conflicting_actions):
        # remove all conflicting options
        for _option_string, action in conflicting_actions:
            action.container._remove_action(action)



from dataclasses import dataclass

import pytest

import draccus
from draccus.utils import Dataclass, DraccusException


@dataclass
class TestConfig:
    a: int = 1
    b: str = "test"
    c: float = 3.14


def test_load_from_file(tmp_path):
    config_path = tmp_path / "config.yaml"
    config_path.write_text("a: 2\nb: hello\nc: 1.23")

    # Test with file path
    cfg = draccus.load(TestConfig, config_path)
    assert cfg.a == 2
    assert cfg.b == "hello"
    assert cfg.c == 1.23

    # Test with file object
    with open(config_path, "r") as f:
        cfg = draccus.load(TestConfig, f)
    assert cfg.a == 2
    assert cfg.b == "hello"
    assert cfg.c == 1.23


def test_loads_from_string():
    config_str = "a: 2\nb: hello\nc: 1.23"
    cfg = draccus.loads(TestConfig, config_str)
    assert cfg.a == 2
    assert cfg.b == "hello"
    assert cfg.c == 1.23


def test_load_backwards_compatibility():
    # Test that load still works with string content directly
    config_str = "a: 2\nb: hello\nc: 1.23"
    cfg = draccus.load(TestConfig, config_str)
    assert cfg.a == 2
    assert cfg.b == "hello"
    assert cfg.c == 1.23


def test_load_invalid_input():
    with pytest.raises(DraccusException):
        draccus.load(TestConfig, None)

    with pytest.raises(DraccusException):
        draccus.loads(TestConfig, None)



import dataclasses
from typing import Optional

from draccus.choice_types import PluginRegistry


@dataclasses.dataclass(frozen=True)
class ModelConfig(PluginRegistry, discover_packages_path="tests.draccus_choice_plugins"):
    layers: int

    @classmethod
    def default_choice_name(cls) -> Optional[str]:
        return "mlp"


@ModelConfig.register_subclass("mlp")
@dataclasses.dataclass(frozen=True)
class MlpConfig(ModelConfig):
    hidden_size: int = 100



import argparse
import dataclasses
import sys

import pytest

from draccus import ParsingError

from .draccus_choice_plugins.model_config import MlpConfig, ModelConfig
from .testutils import TestSetup


def test_plugin_registry_argparse():
    @dataclasses.dataclass(frozen=True)
    class Something(TestSetup):
        model: ModelConfig = MlpConfig(10, 5)

    s = Something.setup("")
    assert s.model == MlpConfig(10, 5)

    s = Something.setup("--model.type mlp --model.layers 12 --model.hidden_size 6")
    assert s.model == MlpConfig(12, 6)

    s = Something.setup("--model.layers 12 --model.hidden_size 6")
    assert s.model == MlpConfig(12, 6)

    s = Something.setup("--model.type gpt --model.layers 12 --model.attn_pdrop 0.2")
    from .draccus_choice_plugins.gpt import GptConfig

    assert s.model == GptConfig(12, 0.2)

    with pytest.raises(argparse.ArgumentError):
        Something.setup("--model.type baby")

    with pytest.raises(ParsingError):
        Something.setup("--model.type gpt --model.layers 12 --model.hidden_size 6")

    with pytest.raises(ParsingError):
        Something.setup("--model.attn_pdrop 12")


# skip this test if using python 3.8
# the help text is a bit different in 3.8


@pytest.mark.skipif(sys.version_info < (3, 10), reason="requires python3.9 or higher")
def test_choice_registry_examine_help():
    @dataclasses.dataclass
    class Something(TestSetup):
        model: ModelConfig = MlpConfig(10, 5)

    # TODO: why is the default: None here?
    target = """
usage: draccus [-h] [--config_path str] [--model str] [--model.type {mlp,gpt}]
               [--model.hidden_size int] [--model.layers int]
               [--model.attn_pdrop float]

options:
  -h, --help            show this help message and exit
  --config_path str     Path for a config file to parse with draccus (default:
                        None)
  --model str           Config file for model (default: None)

test_choice_registry_examine_help.<locals>.Something:

ModelConfig ['model']:

  --model.type {mlp,gpt}
                        Which type of ModelConfig ['model'] to use (default:
                        None)

MlpConfig ['model']:

  --model.layers int
  --model.hidden_size int

GptConfig ['model']:

  --model.layers int
  --model.attn_pdrop float
"""
    print(Something.get_help_text())
    assert Something.get_help_text().strip() == target.strip()
